'''
Created on Dec 28, 2013

@author: Nicklas Boerjesson
'''
from qal.common.strings import make_path_absolute, bool_to_binary_int, string_to_bool

from qal.dataset.custom import CustomDataset
from qal.tools.discover import import_error_to_help


def none_to_zero(_value):
    if _value is None or _value is False:
        return 0
    else:
        return _value


class SpreadsheetDataset(CustomDataset):
    """The Spreadsheet dataset can read data from a spreadsheet, currently only Excel files, and store that data in its."""

    has_header = None
    """True if data begins with a header row containing field names."""
    filename = None
    """The name of the file."""
    field_names = None
    """The names of the fields(if applicable, see has_header)."""
    sheet_name = None
    """The name of the worksheet to use, typically "Sheet1"."""
    x_offset = None
    """The leftmost column where data is held"""
    y_offset = None
    """The topmost row where data is held, *including the header row*."""

    def read_resource_settings(self, _resource):
        self._base_path = _resource.base_path

        if _resource.type.upper() != 'SPREADSHEET':
            raise Exception(
                "SpreadsheetDataset.read_resource_settings.parse_resource error: Wrong resource type: " + _resource.type)
        self.filename = _resource.data.get("filename")
        self.delimiter = _resource.data.get("delimiter")
        if _resource.data.get("has_header"):
            self.has_header = string_to_bool(str(_resource.data.get("has_header")))
        else:
            self.has_header = None

        self.sheet_name = _resource.data.get("sheet_name")

        if _resource.data.get("x_offset") is not None:
            self.x_offset = int(_resource.data.get("x_offset"))
        else:
            self.x_offset = 0
        if _resource.data.get("y_offset") is not None:
            self.y_offset = int(_resource.data.get("y_offset"))
        else:
            self.y_offset = 0

    def write_resource_settings(self, _resource):
        _resource.type = 'SPREADSHEET'
        _resource.data.clear()
        _resource.data["filename"] = self.filename
        _resource.data["delimiter"] = self.delimiter
        _resource.data["has_header"] = self.has_header
        _resource.data["sheet_name"] = self.sheet_name
        _resource.data["x_offset"] = self.x_offset
        _resource.data["y_offset"] = self.y_offset

    def __init__(self, _filename=None, _has_header=None, _resource=None, _sheet_name=None, _x_offset=None,
                 _y_offset=None):
        '''
        Constructor
        '''
        super(SpreadsheetDataset, self).__init__()
        if _resource != None:
            self.read_resource_settings(_resource)
        else:

            if _filename != None:
                self.filename = _filename
            else:
                self.filename = None
            if _has_header != None:
                self.has_header = _has_header
            else:
                self.has_header = None
            if _sheet_name != None:
                self.sheet_name = _sheet_name
            else:
                self.sheet_name = None

            if _x_offset != None:
                self.x_offset = _x_offset
            else:
                self.x_offset = None
            if _y_offset != None:
                self.y_offset = _y_offset
            else:
                self.y_offset = None

    def load_xls(self, _filename, _sheet_name, _x_offset, _y_offset, _header_offset):
        print("SpreadsheetDataset.load_xls: using xlrd")

        try:
            import xlrd
        except ImportError as _err:
            raise Exception(import_error_to_help(_module="xlrd", _err_obj=_err, _pip_package="xlrd",
                                                 _apt_package=None, _win_package=None))
        try:
            _workbook = xlrd.open_workbook(_filename)
            _sheet = _workbook.sheet_by_name(_sheet_name)

            if _header_offset > 0:
                _field_names = _sheet.row_values(rowx=_y_offset, start_colx=_x_offset, end_colx=_sheet.ncols)
            else:
                _field_names = list("Column_" + str(x) for x in range(_y_offset, _sheet.ncols))

            _data_table = []

            for _curr_y in range(_y_offset + _header_offset, _sheet.nrows):
                _data_table.append(_sheet.row_values(_curr_y, start_colx=_x_offset, end_colx=_sheet.ncols))

            return _data_table, _field_names

        except IOError as e:
            raise Exception("SpreadsheetDataset.load: Error reading file:" + str(e))

    def load_xlsx(self, _filename, _sheet_name, _x_offset, _y_offset, _header_offset):
        print("SpreadsheetDataset.load_xlsx: using openpyxl")
        try:
            from openpyxl import load_workbook
        except ImportError as _err:
            raise Exception(import_error_to_help(_module="openpyxl",
                                                 _err_obj=_err,
                                                 _pip_package="openpyxl",
                                                 _apt_package=None,
                                                 _win_package=None))
        try:
            _workbook = load_workbook(_filename)
            _sheet = _workbook.active
            _number_of_columns = _sheet.get_highest_column()
            _number_of_rows = _sheet.get_highest_row()
            _data_table = []
            for _curr_row in _sheet.iter_rows():
                _data_table.append([_cell.value for _cell in _curr_row])

            if _header_offset > 0:
                _field_names = [_sheet.cell(column=_x, row=1 + _y_offset).value for _x in
                                range(1 + _x_offset, _number_of_columns)]
            else:
                _field_names = list("Column_" + str(x) for x in range(1 + _x_offset, _number_of_columns))

            for _curr_y in range(1 + _y_offset + _header_offset, _number_of_rows):
                _data_table.append(
                    [_sheet.cell(column=_x, row=_curr_y).value for _x in range(1 + _x_offset, _number_of_columns)])

            return _data_table, _field_names

        except IOError as e:
            raise Exception("SpreadsheetDataset.load: Error reading file:" + str(e))


    def load(self):
        """Load data. """

        import os.path

        _extension = os.path.splitext(self.filename)[1]

        _x_offset = none_to_zero(self.x_offset)
        _y_offset = none_to_zero(self.y_offset)

        if self.has_header:
            _header_offset = 1
        else:
            _header_offset = 0

        _filename = make_path_absolute(self.filename, self._base_path)

        if _extension.lower() in [".xls", ".xlsx"]:
            self.data_table, self.field_names = self.load_xls(_filename=_filename, _sheet_name=self.sheet_name,
                                                              _x_offset=_x_offset, _y_offset=_y_offset,
                                                              _header_offset=_header_offset)
        # elif _extension.lower() in [".xlsx"]:
        # self.data_table, self.field_names = self.load_xlsx(_filename=_filename, _sheet_name=self.sheet_name,
        #                                    _x_offset=_x_offset, _y_offset=_y_offset,
        #                                    _header_offset = _header_offset)


        elif _extension.lower() == ".odt":
            # from ezodf import newdoc, Paragraph, Heading, Sheet
            raise Exception("SpreadsheetDataset.load: Open Document Spreadsheet not implemented yet.")

        else:
            raise Exception("SpreadsheetDataset.load: Unsupported file type \"" + _extension + "\"")

    def save_xlsx(self, _filename, _data_table, _number_of_rows, _number_of_columns,
                  _x_offset, _y_offset, _header_offset, _field_names, _sheet_name):
        print("SpreadsheetDataset.save_xlsx: using openpyxl")

        try:
            import openpyxl
        except ImportError as _err:
            raise Exception(import_error_to_help(_module="openpyxl",
                                                 _err_obj=_err,
                                                 _pip_package="openpyxl",
                                                 _apt_package=None,
                                                 _win_package=None))

        try:

            _workbook = openpyxl.Workbook()
            _sheet = _workbook.create_sheet(title = _sheet_name)

            if _header_offset > 0:
                for _curr_col_idx in range(0, _number_of_columns):
                    _sheet.cell(row=1 + self.y_offset, column=1 + _curr_col_idx + _x_offset).value = _field_names[_curr_col_idx]

            for _curr_row_idx in range(0, _number_of_rows -1):
                for _curr_col_idx in range(0, _number_of_columns):
                    _sheet.cell(row=1 + _curr_row_idx + _y_offset + _header_offset, column=1 + _curr_col_idx + _x_offset).value = \
                        _data_table[_curr_row_idx][_curr_col_idx]

            _workbook.save(filename=_filename)
        except IOError as e:
            raise Exception("SpreadsheetDataset.save_xlsx: Error saving file:" + str(e))

    def save_xls(self, _filename, _data_table, _number_of_rows, _number_of_columns,
                 _x_offset, _y_offset, _header_offset, _field_names, _sheet_name):
        print("SpreadsheetDataset.save_xls: using xlwt")

        try:
            import xlwt
        except ImportError as _err:
            raise Exception(import_error_to_help(_module="xlwt",
                                                 _err_obj=_err,
                                                 _pip_package="xlwt",
                                                 _apt_package=None,
                                                 _win_package=None))
        try:

            _workbook = xlwt.Workbook()
            _sheet = _workbook.add_sheet(sheetname=_sheet_name, cell_overwrite_ok=True)
            # Write headers
            if _header_offset > 0:
                for _curr_col_idx in range(0, _number_of_columns):
                    _sheet.write(self.y_offset, _curr_col_idx + _x_offset, label=_field_names[_curr_col_idx])
            # Write data
            for _curr_row_idx in range(0, _number_of_rows - 1):
                for _curr_col_idx in range(0, _number_of_columns):
                    _sheet.write(_curr_row_idx + _y_offset + _header_offset, _curr_col_idx + _x_offset,
                                 label=_data_table[_curr_row_idx][_curr_col_idx])

            _workbook.save(filename=_filename)
        except IOError as e:
            raise Exception("SpreadsheetDataset.save_xls: Error saving file:" + str(e))

    def save(self, _save_as=None):
        """Save data"""
        _number_of_rows = len(self.data_table)
        if _number_of_rows > 0:
            _number_of_columns = len(self.data_table[0])
        print("SpreadsheetDataset.save: Filename='" + str(self.filename) + "', Delimiter='" + str(self.delimiter) + "'")
        import os.path

        _extension = os.path.splitext(self.filename)[1]
        if _save_as:
            _filename = _save_as
        else:
            _filename = make_path_absolute(self.filename, self._base_path)

        _x_offset = none_to_zero(self.x_offset)
        _y_offset = none_to_zero(self.y_offset)
        if self.y_offset is None:
            self.y_offset = 0
        if self.has_header:
            _header_offset = 1
        else:
            _header_offset = 0

        if _extension.lower() == ".xls":
            self.save_xls(_filename, _data_table=self.data_table, _number_of_rows=_number_of_rows,
                          _number_of_columns=_number_of_columns, _x_offset=self.x_offset,
                          _y_offset=self.y_offset, _header_offset=_header_offset, _field_names=self.field_names,
                          _sheet_name=self.sheet_name)

        elif _extension.lower() == ".xlsx":
            self.save_xlsx(_filename, _data_table=self.data_table, _number_of_rows=_number_of_rows,
                           _number_of_columns=_number_of_columns, _x_offset=self.x_offset,
                           _y_offset=self.y_offset, _header_offset=_header_offset, _field_names=self.field_names,
                           _sheet_name=self.sheet_name)

        elif _extension.lower() == ".odt":
            # from ezodf import newdoc, Paragraph, Heading, Sheet
            raise Exception("SpreadsheetDataset.load: Open Document Spreadsheet not implemented yet.")

        else:
            raise Exception("SpreadsheetDataset.load: Unsupported file type \"" + _extension + "\"")
        

    
        


"""
Created on Nov 3, 2013

@author: Nicklas Boerjesson
@todo: 
* Add input type checks for better error messages
* Use format string for destination formatting as well (casting from any other primitive type to string)

"""
import re
from lxml import etree
from datetime import date, datetime
from qal.common.strings import empty_when_none
from qal.common.xml_utils import xml_isnone


def perform_transformations(_input, _transformations):
    for _curr_transformation in _transformations:
        _input = _curr_transformation.transform(_input)
    return _input    
    
def make_transformation_array_from_xml_node(_xml_node, _substitution= None):
    _result = []
    for _curr_node in _xml_node:
        if _curr_node.tag.lower() == 'trim':
            _result.append(Trim(_xml_node=_curr_node, _substitution=_substitution))
        elif _curr_node.tag.lower() == 'ifempty':
            _result.append(IfEmpty(_xml_node=_curr_node, _substitution=_substitution))
        elif _curr_node.tag.lower() == 'cast':
            _result.append(Cast(_xml_node=_curr_node, _substitution=_substitution))
        elif _curr_node.tag.lower() == 'replace':
            _result.append(Replace(_xml_node=_curr_node, _substitution=_substitution))
        elif _curr_node.tag.lower() == 'replace_regex':
            _result.append(ReplaceRegex(_xml_node=_curr_node, _substitution=_substitution))
    return _result

def make_transformations_xml_node(_transformations):
    _xml_node = etree.Element("transformations")
    for _curr_transformation in _transformations:
        _xml_node.append(_curr_transformation.as_xml_node())
    
    return _xml_node

class CustomTransformation(object):
    """
    This is the base class for all transformations
    """
    order = None
    """Order dictates when the transformation is run."""
    on_done = None
    """On done is an event, triggered when the transformation has been run. Conveys the resulting value or error message."""
    substitution = None
    """An optional instance of the substitution class. Usually shared by several transformations."""
    def __init__(self, _xml_node = None, _substitution = None):
        """
        Constructor
        """
        if _substitution != None:
            self.substitution = _substitution

        if _xml_node != None:
            self.load_from_xml_node(_xml_node)

    def do_on_done(self, _value=None, _error=None):
        if self.on_done:
            self.on_done(_value, _error)
        return _value
            
    def init_base_to_node(self, _name):
        _xml_node = etree.Element(_name)
        _xml_node.set("order", empty_when_none(self.order))
        return _xml_node
            
    def as_xml_node(self):
        raise Exception("CustomTransformation.as_xml_node : Should not be called. Not implemented in base class, use init_base_to_node().")
        
    def load_from_xml_node(self, _xml_node):
        if _xml_node != None:
            self.order = _xml_node.get("order")
        else:
            raise Exception("CustomTransformation.load_from_xml_node : Base class need a destination node.")

    def transform(self, _value):
        try:
            _result = self._transform(_value)
            return self.do_on_done(_value=_result)
        except Exception as e:
            self.do_on_done(_error="Order: " + str(self.order) + ", " + str(e))
            raise

    def _transform(self, _value):
        """Make transformation"""
        raise Exception("CustomTransformation.transform : Not implemented in base class.")


class Trim(CustomTransformation):
    """Trim returns a copy of the string in which all chars have been trimmed from the beginning and the end of the string (default whitespace characters).
    If the value parameter is set to either "beginning" or "end", only the left or right end of the string is trimmed, respectively."""
    value = None

    def load_from_xml_node(self, _xml_node):
        super(Trim, self).load_from_xml_node(_xml_node)
        self.value = _xml_node.text
    
    def as_xml_node(self):
        _xml_node = self.init_base_to_node("trim")
        _xml_node.text = self.value
        
        return _xml_node

    def _transform(self, _value):
        """Make transformation"""
        if _value is not None:
            if self.value == "beginning":
                return _value.lstrip()
            elif self.value == "end":
                return _value.rstrip()
            else:
                return _value.strip()
        
class IfEmpty(CustomTransformation):
    """IfEmpty returns a specified value if the input value is NULL."""
    value = None
    
    def load_from_xml_node(self, _xml_node):
        super(IfEmpty, self).load_from_xml_node(_xml_node)
        self.value = _xml_node.text
    
    def as_xml_node(self):
        _xml_node = self.init_base_to_node("IfEmpty")
        _xml_node.text =  self.value
        
        return _xml_node

    def _transform(self, _value):
        """Make transformation"""
        if _value is None or _value == "":
            if self.substitution is not None:
                return self.substitution.substitute(self.value)
            else:
                return self.value
        else:
            return _value
        
class Cast(CustomTransformation):
    """ICasts a string to the specified type. The timestamp date format defaults to the ISO format if format_string is not set.\n
    Possible format string directives at : http://docs.python.org/3.2/library/datetime.html#strftime-strptime-behavior\n
    For example, 2013-11-06 22:05:42 is "%Y-%m-%d %H:%M:%S".
    """
    
    dest_type = None
    """The destination type"""
    format_string = None
    """A format string where applicable"""
    
    def load_from_xml_node(self, _xml_node):
        super(Cast, self).load_from_xml_node(_xml_node)
        self.dest_type = xml_isnone(_xml_node.find("dest_type"))
        self.format_string = xml_isnone(_xml_node.find("format_string"))
    
    def as_xml_node(self):
        _xml_node = self.init_base_to_node("cast")
        etree.SubElement(_xml_node, "dest_type").text = self.dest_type
        etree.SubElement(_xml_node, "format_string").text = self.format_string
        
        return _xml_node

    def _transform(self, _value):
        """Make cast"""
        try:
            if _value is None or _value=="":
                return _value
            if self.dest_type in ['string', 'string(255)', 'string(3000)']:
                if isinstance(_value, date):
                    if self.format_string is not None and self.format_string !="":
                        return _value.strftime(self.format_string)
                    else:
                        return _value.strftime("%Y-%m-%d %H:%M:%S")
                else:
                    return str(_value)
            else:
                if isinstance(_value, str):
                    # All other types will not work with quotations
                    _value = re.sub(r'^["\']|["\']$', '', _value)

            if self.dest_type in ['float']:
                return float(_value)
            elif self.dest_type in ['integer', 'serial']:
                return int(_value)
            elif self.dest_type in ['timestamp']:
                if self.format_string is not None and self.format_string !="":
                    return datetime.strptime(_value, self.format_string)
                else:
                    return datetime.strptime(_value, "%Y-%m-%d %H:%M:%S")
                
            elif self.dest_type in ['boolean']:
                return bool(_value)
            else:
                raise Exception("Invalid destination data type: " + str(self.dest_type))
                
        except Exception as e:
            raise Exception("Error in Cast.transform: " + str(e))

class Replace(CustomTransformation):
    """Replace returns a copy of the string in which the occurrences of old have been replaced with new, optionally restricting the number of replacements to max."""
    old = None
    """The old value"""
    new = None
    """The new value"""
    max = None
    """The max number of times to replace"""
    
    def load_from_xml_node(self, _xml_node):
        super(Replace, self).load_from_xml_node(_xml_node)
        self.old = xml_isnone(_xml_node.find("old"))
        self.new = xml_isnone(_xml_node.find("new"))
        self.max = xml_isnone(_xml_node.find("max"))
    
    def as_xml_node(self):
        _xml_node = self.init_base_to_node("replace")
        etree.SubElement(_xml_node, "old").text = self.old
        etree.SubElement(_xml_node, "new").text = self.new
        etree.SubElement(_xml_node, "max").text = self.max
        
        return _xml_node

    def _transform(self, _value):
        """Make replace transformation"""
        # It is a string operation, None will be handled as a string.
        if _value is None:
            _value = ""
        if self.old is None:
            raise Exception("Replace.transform: old value has to have a value.")
        else:
            _old = self.old

        if _value.find(_old) > -1:
            if self.new is None:
                _new = ""
            else:
                _new = self.new

            if self.substitution is not None and _value.find(_old) > -1:
                _new = self.substitution.substitute(_new)

            if self.max:
                return _value.replace(_old, str(_new), int(self.max))
            else:
                return _value.replace(_old, str(_new))
        else:
            return _value

class ReplaceRegex(CustomTransformation):
    """ReplaceRegex returns a copy of the string in which the occurrences of old have been replaced with new, optionally restricting the number of replacements to max."""
    pattern = None
    """The old value"""
    new = None
    """The new value"""
    max = None
    """The max number of times to replace"""
    compiled_regex = None
    """The compiled regular expression"""

    def load_from_xml_node(self, _xml_node):
        super(ReplaceRegex, self).load_from_xml_node(_xml_node)
        self.pattern = xml_isnone(_xml_node.find("pattern"))
        self.new = xml_isnone(_xml_node.find("new"))
        self.max = xml_isnone(_xml_node.find("max"))

    def as_xml_node(self):
        _xml_node = self.init_base_to_node("replace_regex")
        etree.SubElement(_xml_node, "pattern").text = self.pattern
        etree.SubElement(_xml_node, "new").text = self.new
        etree.SubElement(_xml_node, "max").text = self.max

        return _xml_node

    def _transform(self, _value):
        """Make replace transformation"""
        # It is a string operation, None will be handled as a string.
        if _value is None:
            _value = ""
        if (self.pattern is None):
            raise Exception("ReplaceRegex.transform: pattern has to have a value.")
        elif self.compiled_regex is None or self.compiled_regex.pattern != self.pattern:
                self.compiled_regex = re.compile(self.pattern)

        if self.compiled_regex.search(_value):
            if self.new is None:
                _new = ""
            else:
                _new = self.new

            if self.substitution is not None:
                _new = self.substitution.substitute(_new)

            if self.max:
                return self.compiled_regex.sub(_new, _value, int(self.max))
            else:
                return self.compiled_regex.sub(_new, _value)
        else:
            return _value



'''
Created on Sep 21, 2010

@author: Nicklas Boerjesson

'''
from qal.sql.meta import list_class_properties, list_parameter_classes, list_verb_classes, find_class
from qal.sql.types import sql_property_to_type, and_or, \
    constraint_types, index_types, verbs, expression_item_types, \
    condition_part,set_operator, tabular_expression_item_types, data_source_types

from qal.dal.types import db_types
from xml.dom.minidom import Document
from xml.sax.saxutils import escape
from qal.common.xml_utils import XML_Translation, xml_base_type_value, find_child_node, xml_get_text,\
    xml_set_cdata, xml_get_numeric, xml_get_boolean, xml_get_allowed_value, xml_find_non_text_child
from qal.common.resources import Resources

from csv import list_dialects
import os




def sql_property_to_xml_type(_PropertyName):
    """Converts a SQL class' property name to an XML type"""
    result = sql_property_to_type(_PropertyName)
    if result[0] in ['string', 'decimal', 'boolean', 'integer']:
        result[0] = "xsd:" + result[0]
    return result


class SQL_XML(XML_Translation):
    """
    This class converts XML into a class structure(declare in SQL.py) that holds the statements. 
    """


    encoding = 'utf-8'
    _resources = None
    base_path = None
    """The base path of the document. Allows for relative paths within."""


    def __init__(self):
        '''
        Constructor
        '''
        super(SQL_XML, self ).__init__()
        self.namespace = 'http://www.optimalbpm.se/XMLschema/DAL/SQL'
        self.schema_uri = 'http://www.optimalbpm.se/XMLschema/DAL/SQL.xsd'
        self.prefix_own = 'sql'
        self.debuglevel = 2
        self.nestinglevel = 0
    
      
    
    def _add_child_array_of(self, _document, _parentNode, _name, _types):

        _complex_node = _document.createElementNS(self.namespace, self.prefix_schema + ":complexType") 
        _complex_node.setAttribute("name", _name)
        _sequence_node = _document.createElementNS(self.namespace, self.prefix_schema + ":choice") 
        _complex_node.appendChild(_sequence_node)
        for currType in _types: 
            _element_node = _document.createElementNS(self.namespace, self.prefix_schema + ":element")
            _element_node.setAttribute("name", currType)
            _element_node.setAttribute("type", currType)
            _element_node.setAttribute("minOccurs", '0')
            _element_node.setAttribute("maxOccurs", 'unbounded')
            _sequence_node.appendChild(_element_node)
        
        _parentNode.appendChild(_complex_node)
                
    def _add_child_string_restriction(self, _document, _parent_node, _name, _enums, _pattern = None):
        _simple_node = _document.createElementNS(self.namespace, self.prefix_schema + ":simpleType") 
        _simple_node.setAttribute("name", _name)
        _restriction_node = _document.createElementNS(self.namespace, self.prefix_schema + ":restriction") 
        _restriction_node.setAttribute("base", self.prefix_schema + ":string")
        if _enums != None: 
            for _curr_value in _enums:
                _enumeration_node = _document.createElementNS(self.namespace, self.prefix_schema + ":enumeration") 
                _enumeration_node.setAttribute("value", _curr_value)
                _restriction_node.appendChild(_enumeration_node)
        if _pattern != None:        
            _pattern_node = _document.createElementNS(self.namespace, self.prefix_schema + ":pattern") 
            _pattern_node.setAttribute("value", _pattern)
            _restriction_node.appendChild(_pattern_node)
        
        _simple_node.appendChild(_restriction_node)  
        _parent_node.appendChild(_simple_node)      

    def _add_child_type_restriction(self, _document, _parent_node, _name, _enums):
        _simple_node = _document.createElementNS(self.namespace, self.prefix_schema + ":complexType") 
        _simple_node.setAttribute("name", _name)
        _choice_node = _document.createElementNS(self.namespace, self.prefix_schema + ":choice") 
        for _curr_value in _enums:
            _enumeration_node = _document.createElementNS(self.namespace, self.prefix_schema + ":element") 
            _enumeration_node.setAttribute("name", _curr_value)
            _enumeration_node.setAttribute("type", _curr_value)
            _choice_node.appendChild(_enumeration_node)
            

        _simple_node.appendChild(_choice_node)  
        _parent_node.appendChild(_simple_node)     
          
    def _add_child_restrictions(self, _document, _parent_node): 
        self._add_child_string_restriction(_document, _parent_node, "datatypes", None, "(integer|string|string(\(.*\))|serial|timestamp)")   
        self._add_child_string_restriction(_document, _parent_node, "db_types", db_types())   
        self._add_child_string_restriction(_document, _parent_node, "and_or", and_or())   
        self._add_child_string_restriction(_document, _parent_node, "index_types", index_types())   
        self._add_child_string_restriction(_document, _parent_node, "constraint_types", constraint_types())
        self._add_child_string_restriction(_document, _parent_node, "set_operator", set_operator())
        self._add_child_string_restriction(_document, _parent_node, "data_source", data_source_types())
        self._add_child_string_restriction(_document, _parent_node, "csv_dialects", list_dialects())
                      
        self._add_child_type_restriction(_document, _parent_node, "statement", verbs())    
        self._add_child_type_restriction(_document, _parent_node, "condition_part", condition_part()) 
        self._add_child_type_restriction(_document, _parent_node, "tabular_expression_item", tabular_expression_item_types()) 
                    
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterString', ['ParameterString'])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterConstraint', ['ParameterConstraint'])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterColumndefinition', ['ParameterColumndefinition'])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterSource', ['ParameterSource'])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterWhen', ['ParameterWhen'])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterIdentifier', ["ParameterIdentifier"])
        self._add_child_array_of(_document, _parent_node, 'Array_Statement', ["statement"])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterOrderByItem', ["ParameterOrderByItem"])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterCondition', ["ParameterCondition"])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterField', ["ParameterField"])
        self._add_child_array_of(_document, _parent_node, 'Array_ParameterAssignment', ["ParameterAssignment"])
        self._add_child_array_of(_document, _parent_node, 'Array_expression_item', expression_item_types())
        self._add_child_array_of(_document, _parent_node, 'Array_tabular_expression_item', tabular_expression_item_types())    
        self._add_child_array_of(_document, _parent_node, 'Array_list', '*')        

        
                
    def _add_child_property_node(self, _document, _parent_node, _property_name):
        _curr_node = _document.createElementNS(self.namespace, self.prefix_schema + ":element")
        _curr_node.setAttribute("name", _property_name)

        _curr_node.setAttribute("type", sql_property_to_xml_type(_property_name)[0])
        _curr_node.setAttribute("minOccurs", "0")
        _parent_node.appendChild(_curr_node)

    def _add_child_type_node(self, _document, _parent_node, _class_name):
        
        _complex_node = _document.createElementNS(self.namespace, self.prefix_schema + ":complexType") 
        _complex_node.setAttribute("name", _class_name)

        _all_node = _document.createElementNS(self.namespace, self.prefix_schema + ":all") 
        _complex_node.appendChild(_all_node)
        
        for _curr_property in list_class_properties(_class_name):
            self._add_child_property_node(_document, _all_node, _curr_property)  
                      
        _parent_node.appendChild(_complex_node)
              
     
    def _add_child_types(self, _document, _parentNode):
           
        #First add restrictions
            
        self._add_child_restrictions(_document, _parentNode)       
        #First, Add parameter types
        for _curr_class in list_parameter_classes():
            self._add_child_type_node(_document, _parentNode, _curr_class)
           
        #Then add verbs.
            
        for _curr_class in list_verb_classes():
            self._add_child_type_node(_document, _parentNode, _curr_class)
            
    
    def generate_schema(self):
        """Generates an XML schema based on the class structure in SQL.py"""
        #Create the minidom document
        doc = Document();
        
        # Create the <wml> base element
        schema = doc.createElement(self.prefix_schema + ":schema")
        schema.setAttribute("xmlns:" + self.prefix_schema, 'http://www.w3.org/2001/XMLSchema')
        schema.setAttribute("targetNamespace", self.namespace)
        schema.setAttribute("xmlns", self.namespace)
        schema.setAttribute("elementFormDefault", 'qualified')
        doc.appendChild(schema) 
     
        self._add_child_types(doc, schema);
        statement = doc.createElement(self.prefix_schema + ":element")
        statement.setAttribute("name", 'statement')
        statement.setAttribute("type", 'statement')
        schema.appendChild(statement)
        
        return doc
    
    
    def _parse_array_xml_node(self, _node, _list, _parent):
        self._go_down("_parse_array_xml_node") 
        self._debug_print("_parse_array_xml_node: Parsing " + _node.nodeName)
        
        # Loop nodes and parse them.        
 
        for currNode in _node.childNodes:
            
            if currNode.nodeType != currNode.TEXT_NODE:
                # Do not handle text nodes, there should never be any in lists.
                resobj = self._parse_class_xml_node(currNode, None, _parent)
                _list.append(resobj)
                
        self._get_up("_parse_array_xml_node")    
        return _list
    
    def _parse_class_xml_node(self, _node,_classname, _parent):
        self._go_down("_parse_class_xml_node")
        if _classname == None:
            _classname = _node.nodeName
        
                                      
        self._debug_print("_parse_class_xml_node: Parsing " + _classname)

        _stripped_classname = self._strip_own(_classname)
        
        # Check for base typesError in VerbCreateIndex, name is not set.
        if _stripped_classname.lower() in  ['str', 'int', 'float', 'datetime']:
            return xml_base_type_value(_node, _stripped_classname) 
        
        # Find and instatiate the actual class.
        _obj, _obj_name = find_class(_stripped_classname)
        
        if hasattr(_obj, 'as_sql'):
            _obj._parent = _parent
            _obj._base_path = self.base_path
            self._debug_print("_parse_class_xml_node: Found matching Parameter class for " + _classname + " : " + _obj_name)

        elif isinstance(_obj, list):
            # If this is a list, parse it and return.
            self._debug_print("_parse_class_xml_node: Found matching list class for " + _classname + " : " + _obj_name)
            return self._parse_array_xml_node(_node, _obj, _obj) 
        else:
            raise Exception("_parse_class_xml_node: Could not find matching class : " + _obj_name)
        
        # Loop the object's properties
        for _curr_itemkey, _curr_obj in _obj.__dict__.items():

            if _curr_itemkey != 'row_separator':
                
                _curr_node = find_child_node(_node, self._add_own(_curr_itemkey))
                if _curr_node != None:
                    
                    self._debug_print("_parse_class_xml_node: Parsing property " + _curr_itemkey)
                    
                    if isinstance(_curr_obj, list):
                        _obj.__dict__[_curr_itemkey] = self._parse_array_xml_node(_curr_node, _curr_obj, _obj)
                    else:
                        # Match the property to a type.
                        currtype = sql_property_to_type(_curr_itemkey)
                        if currtype[0].lower() in ['string', 'datatypes']:
                            _obj.__dict__[_curr_itemkey] = xml_get_text(_curr_node)
                        elif currtype[0].lower() == 'boolean':
                            _obj.__dict__[_curr_itemkey] = xml_get_boolean(_curr_node)
                        elif currtype[0].lower() in ['integer', 'float', 'decimal']:
                            _obj.__dict__[_curr_itemkey] = xml_get_numeric(_curr_node, currtype[0])
                        elif currtype[0:5] == 'verb_' or currtype[0:10] == 'parameter_':
                            raise Exception("_parse_class_xml_node: Strange VERB/PARAMETER happened parsing.. parent: "+ _parent.__class__.__name__ + "Class: " +_classname+" Currtype: " +currtype)
                            _obj.__dict__[_curr_itemkey] = self._parse_class_xml_node(_curr_node, _obj, _obj)
                           
                        elif len(currtype) > 1 and type(currtype[1]) == list:
                            _curr_child = xml_find_non_text_child(_curr_node)
                            if (_curr_child):
                                _obj.__dict__[_curr_itemkey] = self._parse_class_xml_node(_curr_child, None, _obj)    
                            else:
                                # Base types doesn't have any children.
                                _obj.__dict__[_curr_itemkey] = xml_get_allowed_value(_curr_node, currtype)
                                
                        else:
                            _curr_child = xml_find_non_text_child(_curr_node)
                            if (_curr_child):
                                _obj.__dict__[_curr_itemkey] = self._parse_class_xml_node(_curr_node, currtype[0], _obj)
               
                  
        if self._resources and hasattr(_obj, 'resource_uuid') and _obj.resource_uuid != None and _obj.resource_uuid != '':
            _obj._resource = self._resources.get_resource(_obj.resource_uuid)
            self._debug_print("_parse_class_xml_node: Added resource_uuid for " + _obj_name + ": " + _obj.resource_uuid, 1)
        
        self._get_up("_parse_class_xml_node")       
        return _obj

    def xml_to_sql_structure(self, _xml = "", _node = None, _base_path = None):
        """Translates an XML file into a class structure"""
        _node = self.get_root_node('statement', _xml, _node)

        if _base_path:
            self.base_path = os.path.dirname(_base_path)

        _resources_node = find_child_node(_node, 'resources')


        if _resources_node:  
            self._debug_print("xml_to_sql_structure: Found resources.")
            # Send XML here, since resources now uses lxml      
            self._resources = Resources(_resources_xml= _resources_node.toxml(), _base_path = _base_path)
        
        _verb = xml_find_non_text_child(_node)
        if (_verb == None):
            raise Exception('XMLToSQL: No Verb_*-node found.') 
        _structure = self._parse_class_xml_node(_verb, None, None)        

                
        return _structure
    
    def xml_file_to_sql(self, _xml_file_name, **kwargs):
        """Reads a specified XML fil and translates it into an SQL class structure.""" 
        # Read file
        _xml_file = open(_xml_file_name, 'r')
        _xml = _xml_file.read()
        # Encode parameters
        for _curr_name, _curr_value in kwargs.items():
            if _curr_value == None:
                _curr_value = 'NULL'
            _xml = _xml.replace('::Param='+ _curr_name + '::', escape(str(_curr_value)))
        _xml_file.close()
        return self.xml_to_sql_structure(_xml)
    
    def _xml_encode_list(self, _document, _node, _list):
        self._debug_print("_xml_encode_list: Encoding " + _node.nodeName)
        for _curr_item in _list:
            self._xml_encode_object(_document, _node, _curr_item)

        
            
    
    def _xml_encode_object(self, _document, _parent_node, _object):
        
        if (_object != None and _object != ""):
            self._debug_print("_xml_encode_object: Encoding " + _object.__class__.__name__)
            _object_node = _document.createElement(self.prefix_own + ':' + _object.__class__.__name__)
            _parent_node.appendChild(_object_node)                
            if hasattr(_object, "__dict__"):
                for _curr_property_name, currProperty in sorted(_object.__dict__.items()):
                    if not _curr_property_name.lower() in ['row_separator'] and not hasattr(currProperty, '__call__') and _curr_property_name[0:1] != '_':
                        # Create node for property
                        _curr_node = _document.createElement(self.prefix_own + ':' + _curr_property_name)
                               
                        # Property is a list
                        if isinstance(currProperty, list):
                            self._xml_encode_list(_document, _curr_node, currProperty)
                        elif hasattr(currProperty, 'as_sql'):
                            self._xml_encode_object(_document, _curr_node, currProperty)
                        else:
                            _curr_type = sql_property_to_type(_curr_property_name)     
                            if str(currProperty).isnumeric() and len(_curr_type) > 1:
                                xml_set_cdata(_curr_node, _curr_type[1][currProperty])
                            else:
                                xml_set_cdata(_curr_node, currProperty, (_curr_type == "boolean"))
                            
                        _object_node.appendChild(_curr_node)
                        
            elif isinstance(_object, list): 
    
                self._xml_encode_list(_document, _object_node, _object)
            else:
                xml_set_cdata(_object_node, _object) 

    
    def sql_structure_to_xml(self, _structure):
        """Translates an XML structure into XML"""
        #Create the minidom document
        _doc = Document();
        _doc.encoding = self.encoding
        # Create the root element "statement".
        _statement = _doc.createElement(self.prefix_own + ":statement")
        _statement.setAttribute("xmlns:" + self.prefix_xml, 'http://www.w3.org/2001/XMLSchema')
        _statement.setAttribute(self.prefix_xml + ":schemaLocation", self.namespace + ' ' + self.schema_uri)
        _statement.setAttribute("xmlns:" + self.prefix_own, self.namespace)
        _doc.appendChild(_statement) 
        
        # Recurse structure
        self._xml_encode_object(_doc, _statement, _structure);
        
        return _doc


        
        
        



#!/usr/bin/python3.3

'''
Created on Aug 11, 2013

@author: Nicklas Boerjesson
'''

from setuptools import setup

import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..', 'source'))
sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..', 'source', 'qal'))

from qal import __release__

if __name__ == "__main__":
    setup(
        name='qal',
        version=__release__,
        description='QAL is a Python library for mixing and merging data involving different sources and destinations.',
        author='Nicklas Boerjesson',
        author_email='nicklas_attheold_optimalbpm.se',
        maintainer='Nicklas Boerjesson',
        maintainer_email='nicklas_attheold_optimalbpm.se',
        long_description="""QAL is a Python library for mixing and merging data involving different sources and destinations.\n
        It supports several database backends and file formats.
          """,
        url='https://sourceforge.net/projects/qal/',
        packages=['qal', 'qal.dal', 'qal.sql',
                  'qal.common', 'qal.dataset', 'qal.tools'],
        package_data = {
            # If any package contains *.txt or *.xml files, include them:
            '': ['*.txt', '*.xml', '*.sql']
        },
        license='BSD')



"""
Created on Aug 17, 2010

@author: Nicklas Boerjesson
"""

from qal.common.listhelper import CI_index

DB_MYSQL = 0
DB_POSTGRESQL = 1
DB_ORACLE = 2
DB_DB2 = 3
DB_SQLSERVER = 4


def unenumerate(value, _Type):
    """Returns the value of a specific item"""
    return value[_Type]   

def db_types():
    """Returns a list of supported database engines"""
    return ['MySQL', 'PostgreSQL', 'Oracle','DB2','SQLserver'];

def string_to_db_type(_value):
    """Returns db_type constant matching the specified string"""
    result = CI_index(db_types(),_value)
    if result > -1:
        return result
    else:
        raise Exception("string_to_db_type: Invalid database type:" + str(_value)) 

        
    
def db_type_to_string(_DBType):
    """Returns string matching the specified db_type constant"""""
    try:
        return unenumerate(db_types(), _DBType)
    except:
        raise Exception("db_type_to_string: Invalid database type:" + str(_DBType))



"""
Created on Oct 3, 2010

@author: Nicklas Boerjesson
@note: This module contains functionality to retrieve metadata about databases and their objects.
@todo: These SQL:s could be moved into an XML file(VerbCustom).
The gain is yet somewhat marginal. They should not move in to the database due to versioning.  
"""
from builtins import classmethod

_table_info_postgresql = "SELECT a.attname AS COLUMN_NAME, t.typname AS DATA_TYPE,\
       CASE WHEN a.attlen > -1 THEN a.attlen  ELSE a.atttypmod END - 4 AS DATA_LENGTH, NOT a.attnotnull AS NULLABLE \
FROM pg_class c, pg_attribute a, pg_type t \
WHERE c.relname = ':TABLENAME'\
  AND a.attnum > 0\
  AND a.attrelid = c.oid\
  AND a.atttypid = t.oid \
ORDER BY a.attnum;"
  
_table_info_mysql = "SELECT COLUMN_NAME, DATA_TYPE, IS_NULLABLE AS NULLABLE, CHARACTER_MAXIMUM_LENGTH AS DATA_LENGTH \
FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=':DATABASENAME' AND TABLE_NAME = ':TABLENAME'"

_table_info_oracle = "SELECT COLUMN_NAME, DATA_TYPE, NULLABLE, DATA_LENGTH FROM user_tab_columns \
WHERE table_name=':TABLENAME'"

_table_info_db2 = "SELECT NAME AS COLUMN_NAME, COLTYPE AS DATA_TYPE, NULLS AS NULLABLE, LENGTH AS DATA_LENGTH \
FROM SYSIBM.SYSCOLUMNS WHERE UPPER(TBNAME) = UPPER(':TABLENAME');"

_table_info_sqlserver = "SELECT \
clmns.name AS [COLUMN_NAME],\
usrt.name AS [DATA_TYPE],\
clmns.is_nullable AS [NULLABLE],\
usrt.max_length AS [DATA_LENGTH]\
FROM \
sys.tables AS tbl \
INNER JOIN sys.all_columns AS clmns ON clmns.object_id=tbl.object_id \
LEFT OUTER JOIN sys.types AS usrt ON usrt.user_type_id = clmns.user_type_id \
WHERE \
(tbl.name=':TABLENAME') \
ORDER BY clmns.column_id ASC"

_table_list_mysql_by_database_name =  _table_list_oracle_by_schema = table_list_sqlserver = ""


_table_list_postgresql_by_database_name = "SELECT  table_name \
FROM information_schema.tables \
WHERE table_type = 'BASE TABLE' \
    AND table_schema = 'public' \
    AND table_catalog = ':DATABASE' \
ORDER BY table_type, table_name "

_table_list_mysql_by_database_name = "SHOW TABLES IN :DATABASE"
_table_list_oracle_by_database_name = "select table_name from dba_tables;"
_table_list_db2_by_database_name = "SELECT NAME FROM SYSIBM.SYSTABLES WHERE Type = 'T';"
_table_list_mssql_by_database_name = "SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES"

_table_list_db2_by_schema = "select NAME from sysibm.systables \
where CREATOR  = ':USER' \
and type = 'T';"

_table_list_oracle_by_schema = "select table_name from user_tables WHERE TABLESPACE_NAME = ':USER' AND table_name NOT LIKE 'DEF$_%'"

def _table_info_sql(_db_type, _table_name, _database_name):
    """Lists information about a table"""
    SQL = [_table_info_mysql, _table_info_postgresql, _table_info_oracle, _table_info_db2, _table_info_sqlserver][_db_type]
    SQL = SQL.replace(":TABLENAME", _table_name)
    SQL = SQL.replace(":DATABASENAME", _database_name)
    return SQL

def _table_list_sql_by_schema(_db_type, _user):
    """List tables in a schema"""
    if not(_db_type in [2,3]):
        raise Exception("table_list_sql_by_schema: Only DB2 and Oracle supported currently!") 
    
    SQL = [_table_list_mysql_by_database_name, _table_list_postgresql_by_database_name, _table_list_oracle_by_schema, _table_list_db2_by_schema, table_list_sqlserver][_db_type]
    SQL = SQL.replace(":USER", _user)
    return SQL

def _table_list_sql_by_database_name(_db_type, _database):
    """List tables in a specified database"""
    if not(_db_type in [0,1,3,4]):
        raise Exception("table_list_sql_by_database_name: Only MySQL, Postgres, DB2 and SQL Server is supported currently!")
    
    SQL = [_table_list_mysql_by_database_name, _table_list_postgresql_by_database_name, _table_list_oracle_by_database_name, _table_list_db2_by_database_name, _table_list_mssql_by_database_name][_db_type]
    SQL = SQL.replace(":DATABASE", _database)
    return SQL


class Meta_Queries(object):
    """The meta queries class collects methods for gathering meta data about a database."""

    @staticmethod
    def table_info(_dal, _table_name):
        """
        List columns in the specified table
        :param _table_name: The name of the table to list columns for
        :return: A list of column names
        """
        if not _dal.connected:
            _dal.connect_to_db()
        _rows = _dal.query(_table_info_sql(_dal.db_type, _table_name, _dal.db_databasename))
        columns = []
        for row in _rows:
            columns.append(row[0])
        return columns   
     
    @staticmethod
    def table_list_by_schema(_dal, _schema_name):
        """
        List tables in the specified schema
        :param _dal: A connected instance of DatabaseAbstractionLayer (see qal.dal)
        :param _schema_name: A schema name
        :return: A list of table names
        """
        if not _dal.connected:
            _dal.connect_to_db()
        rows = _dal.query(_table_list_sql_by_schema(_dal.db_type, _schema_name))
        columns = list()
        for row in rows:
            columns.append(row[0])
        return columns    

    @staticmethod
    def table_list_by_database_name(_dal, _database_name):
        """
        List tables in the specified database
        :param _dal: A connected instance of DatabaseAbstractionLayer (see qal.dal)
        :param _database_name: A database name
        :return: A list of table names
        """
        if not _dal.connected:
            _dal.connect_to_db()

        """TODO: Refactor so either MySQL gets connection's name or all else goes by database name(which won't work)"""
        rows = _dal.query(_table_list_sql_by_database_name(_dal.db_type, _database_name))
        print("SQL:\n"+ _table_list_sql_by_database_name(_dal.db_type, _database_name) + "\n_database_name:" +
              _database_name + "\nrows: \n" + str(rows))
        columns = list()
        for row in rows:
            columns.append(row[0])
        return columns

    @staticmethod
    def oracle_all_sequences(_dal):
        """
        Oracle specific: List all sequences in an Oracle database
        :param _dal: A connected instance of DatabaseAbstractionLayer (see qal.dal)
        :return: A list or sequence names
        """
        if not _dal.connected:
            _dal.connect_to_db()
        rows = _dal.query("SELECT SEQUENCE_NAME FROM USER_SEQUENCES")
        sequences = list()
        for row in rows:
            sequences.append(row[0])
        return sequences    
        


"""
    *********************************************
    qal.sql is an abstraction of the SQL language
    ********************************************* 
    * Abstracts SQL to enable a single object structure to platform-independently generate SQL for each platform.
    * Can translate itself from and to an XML structure for portability and persistence.
    * Can implicitly join any qal.dataset from all kinds of sources directly on a target server(loads them into temp tables and joins then).
    * Currently available qal.dataset:s : flat file, matrix, XPath(XML/HTML/XHTML) and spreadsheet(Excel-formats/.odt).      
    * RBDMS Platforms (qal.dal.types) : MySQL, PostgreSQL, Oracle, DB2 and SQL server
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""


"""
    Contains miscellaneous string functions.
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details. 
"""

import os

def parse_balanced_delimiters(_input, _d_left, _d_right, _text_qualifier):
    """Removes all balanced delimiters. Handles multiple levels and ignores those contained within text delimiters."""
    
    # Depth is deep into recursion we are
    _depth = 0
    
    # Balance is a list of all balanced delimiters
    _balanced = []
    
    # Cleared is a string without the balanced delimiters
    _cleared = ""
    
    # If in text mode all delimiters are ignored, since they are in text. Also, _text delimiters are escaped.
    _text_mode = False
    
    # Start- and end positions refers to start and end of balanced delimiters. 
    _start_pos = None
    # End pos implicitly also mean start of "clean" area.
    _clear_start_pos = 0
    
    for _curr_idx in range(len(_input)):
        _curr_char = _input[_curr_idx]


        if _curr_char == _text_qualifier:
            _text_mode = not _text_mode
        elif _text_mode == False:
            if _curr_char == _d_left:
                if _depth == 0:
                    _start_pos = _curr_idx
                    _cleared+=_input[_clear_start_pos:_curr_idx]
                    _clear_start_pos = None
                    
                _depth+=1
            elif _curr_char == _d_right:
                _depth-=1
            if _start_pos and _depth == 0  and _start_pos < _curr_idx:
                _balanced.append(_input[_start_pos +1:_curr_idx])
                _start_pos = None
                _clear_start_pos = _curr_idx + 1 
                
            
    if _cleared == "":
        return _balanced, _input
    else:
        return _balanced, _cleared

def empty_if_none(_string, _source = None):
    """If _source if None, return an empty string, otherwise return string. 
    This is useful when you build strings and want it to be empty if you don't have any data.
    For example when building a comma-separated string with a dynamic number of parameters: 
    .. code-block:: python
    
    command = (_id +";" +  +  empty_if_none(";"+str(_comment), _comment))
    """
    if _source is None:
        return ""
    else:
        return str(_string)

def empty_when_none(_string = None):
    """If _string if None, return an empty string, otherwise return string.
    """
    if _string is None:
        return ""
    else:
        return str(_string)


def make_path_absolute(_path, _base_path):
    """Makes a path defined as relative in the XML resource definition absolute using _base_path."""
    _path = os.path.normpath(_path)
    if _base_path is not None and _base_path != "" :
        _base_path = os.path.normpath(_base_path)
    if  os.path.isabs(_path):
        # Don't do anything, path is already absolute.
        return _path
    elif _base_path:
        return os.path.join(_base_path, _path)
    else:
        raise Exception("Resource.make_path_absolute: make_path_absolute cannot make " + _path + " absolute without a _base_path.")

def bool_to_binary_int(_value):
    """Converts True to 1 and False and None to 0."""
    if _value is None or _value is False:
        return 0
    elif _value is True:
        return 1


def binary_int_to_bool(_value):
    """Converts 1 to True and 1 and None to False."""
    if _value is None or _value==0:
        return False
    elif _value==1:
        return True

def string_to_bool(_value):
    """Converts "True" to True None to False. Case-insensitive."""
    if _value is None or _value.lower() == "true":
        return True
    else:
        return False



'''
Created on Sep 14, 2012

@author: Nicklas Boerjesson
'''
from datetime import date
import datetime
from itertools import islice
from qal.common.strings import make_path_absolute, string_to_bool

from qal.dataset.custom import CustomDataset

import csv
from urllib.parse import unquote


class FlatfileDataset(CustomDataset):
    """This class loads a flat file into an array, self.data_table."""
    delimiter = None
    """The delimiter, typically ";", "," or "|"."""
    has_header = None
    """True if data begins with a header row containing field names."""
    filename = None
    """The name of the file"""
    field_names = None
    """The names of the fields(if applicable, see has_header)"""
    csv_dialect = None
    """Specifies the csv dialect"""
    quoting = None
    """To what degree does the file employ quoting, values:
     * None = No quoting at all, quotes will be treated as values
     * "MINIMAL" = Quoting is only used when necessary
     * "ALL" = Quoting is applied used on all fields.
     * "NONNUMERIC" = Non-numeric fields are quoted."""
    quotechar = "\""
    """What character to use for quoting. Normally "\\"\"."""
    escapechar = None
    """What character is used to escape special characters, typically "\\"."""
    lineterminator = None
    """What character is used to indicate the end of a line. typically "\\n"."""
    skipinitialspace = None
    """True if initial spaces should be disregarded."""
    
    def __init__(self, _delimiter = None, _filename = None, _has_header = None, _csv_dialect = None, _resource = None, _quoting = None, _quotechar = None, _skipinitialspace = None):
        """Constructor"""
        super(FlatfileDataset, self ).__init__()
       
        if _resource != None:
            self.read_resource_settings(_resource)
        else:
            if _delimiter != None: 
                self.delimiter = _delimiter
            else:  
                self.delimiter = None    
            if _filename != None: 
                self.filename = _filename
            else:
                self.filename = None      
            if _has_header != None: 
                self.has_header = _has_header
            else:
                self.has_header = None
                  
            if _csv_dialect != None: 
                self.csv_dialect = _csv_dialect
            else:
                self.csv_dialect = None      
             
            if _quoting != None: 
                self.quoting = _quoting
            else:
                self.quoting = None

            if _quotechar != None:
                self.quotechar = _quotechar
            else:
                self.quotechar = None

            if _skipinitialspace != None:
                self.skipinitialspace = _skipinitialspace
            else:
                self.skipinitialspace = None
        
    def read_resource_settings(self, _resource):
        if _resource.type.upper() != 'FLATFILE':
            raise Exception("FlatfileDataset.read_resource_settings.parse_resource error: Wrong resource type: " + _resource.type)
        self._base_path = _resource.base_path
        self.filename = _resource.data.get("filename")
        self.delimiter = _resource.data.get("delimiter")
        if _resource.data.get("has_header"):
            self.has_header = string_to_bool(str(_resource.data.get("has_header")))
        else:
            self.has_header = None
        self.csv_dialect = _resource.data.get("csv_dialect")
        self.quoting = _resource.data.get("quoting")
        self.escapechar = _resource.data.get("escapechar")
        if _resource.data.get("lineterminator") is not None:
            self.lineterminator = bytes(_resource.data.get("lineterminator"), "UTF-8").decode("unicode-escape")
        else:
            self.lineterminator = None
        self.quotechar = _resource.data.get("quotechar") or '"'       
        self.skipinitialspace = _resource.data.get("skipinitialspace")

    def write_resource_settings(self, _resource):
        # Clear first, one could be overwriting an resource with other data fields
        _resource.data = {}
        _resource.type = _resource.type
        _resource.data["filename"] = self.filename
        _resource.data["delimiter"] = self.delimiter
        _resource.data["has_header"] = self.has_header
        _resource.data["csv_dialect"] = self.csv_dialect
        _resource.data["quoting"] = self.quoting
        _resource.data["escapechar"] = self.escapechar
        _resource.data["lineterminator"] = self.lineterminator
        _resource.data["quotechar"] = self.quotechar
        _resource.data["skipinitialspace"] = self.skipinitialspace
         
    def _quotestr_to_constants(self, _str):
        if _str == None:
            return csv.QUOTE_NONE
        elif _str.upper() == "MINIMAL":
            return csv.QUOTE_MINIMAL
        elif _str.upper() == "ALL":
            return csv.QUOTE_ALL
        elif _str.upper() == "NONNUMERIC":
            return csv.QUOTE_NONNUMERIC
        else:
            raise Exception("Error in _quotestr_to_constants: " + str(_str) + " is an invalid quotestr.")
            

    def load(self):
        """Load data"""
        print("FlatfileDataset.load: Filename='" + str(self.filename) + "', Delimiter='"+str(self.delimiter)+"'" +
              ", Base_path "+ str(self._base_path))
        
        _file = open( make_path_absolute(self.filename, self._base_path), 'r')
        _reader = csv.reader(_file, 
                             delimiter=self.delimiter, 
                             quoting=self._quotestr_to_constants(self.quoting),
                             quotechar = self.quotechar,
                             skipinitialspace = self.skipinitialspace
                             )
        _first_row = True
        self.data_table = []
        for _row in _reader:
            # Save header row if existing.
            if (_first_row):
                if self.has_header == True:
                    self.field_names = [_curr_col.replace("'", "").replace("\"", "") for _curr_col in _row]
                    print("self.field_names :" + str(self.field_names))
                else:
                    self.field_names = []
                    for _curr_idx in range(0,len(_row)):
                        self.field_names.append("Field_"+ str(_curr_idx))

                _first_row = False
            else:
                self.data_table.append(_row)

        _file.close()
        return self.data_table   

    def save(self, _save_as = None):
        """Save data"""

        print("FlatfileDataset.save: Filename='" + str(self.filename) + "', Delimiter='"+str(self.delimiter)+"'")
        
        if _save_as:
            _filename = _save_as
        else:
            _filename =  make_path_absolute(self.filename, self._base_path)
            
        _file = open(_filename, 'w')

        # Work around silly line terminator default, has to be set if passed.
        if self.lineterminator:
            _writer = csv.writer(_file,                              
                             delimiter=self.delimiter, 
                             quoting=self._quotestr_to_constants(self.quoting),
                             quotechar = self.quotechar,
                             skipinitialspace = self.skipinitialspace,
                             lineterminator = self.lineterminator
                             )
        else:
            _writer = csv.writer(_file,                              
                             delimiter=self.delimiter, 
                             quoting=self._quotestr_to_constants(self.quoting),
                             quotechar = self.quotechar,
                             skipinitialspace = self.skipinitialspace
                             )

            
        if self.has_header == True:
            _writer.writerow(self.field_names)
    
            
        _writer.writerows(self.data_table)
            
        _file.close()


"""
    Variable substitution functionality
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""

import datetime
import uuid
from getpass import getuser

class Substitution(object):
    """The Substitution class is responsible for replacing and generating variables for a substitution session.
    Using a session one can, for example, keep track of an identity increment.
    """
    builtin_substitutions = None

    def __init__(self):

        self.builtin_substitutions = {
            "identity": self._builtin_identity,
            "uuid": self._builtin_uuid,
            "username": self._builtin_username,
            "curr_datetime": self._builtin_curr_datetime
        }
        self._builtin_identity_value = None

    def _builtin_username(self):
        """TODO: Secure this, this value depends on environment variables. On the other hand, the users doesn't need
        to use this function to insert a false user name.
        """
        return getuser()

    def _builtin_curr_datetime(self):
        return datetime.datetime.now()

    def _builtin_uuid(self):
        return uuid.uuid4()

    def _builtin_identity(self):
        """Returns an int identity value"""
        if self._builtin_identity_value is None:
            self._builtin_identity_value = 0

        _result = self._builtin_identity_value
        self._builtin_identity_value += 1
        return _result

    def set_identity(self, _value):
        self._builtin_identity_value = int(_value)

    def substitute(self, _input):
        """ The replace method scans _input for known substitution variables.
        It tries to keep the value within the same data type.
        :param _input: The substitution string, potentially holding substitution variables
        :return: Returns a value where all substitutions are made
        """

        for _var_name, _var_func in self.builtin_substitutions.items():
            _curr_subst = "::" + _var_name + "::"
            if _input == _curr_subst:
                # Its a single value, return value immidiately
                return _var_func()
            else:
                if _input.find(_curr_subst) > -1:
                    _input = _input.replace(_curr_subst, str(_var_func()))

        return _input



"""
Created on Sep 23, 2013

@author: Nicklas Boerjesson
"""

from datetime import datetime
from qal.dal.types import DB_MYSQL, DB_POSTGRESQL
"""
MySQL constants, made from: https://github.com/PyMySQL/PyMySQL/blob/master/pymysql/constants/FIELD_TYPE.py
"""
MySQL_DECIMAL = 0
MySQL_TINY = 1
MySQL_SHORT = 2
MySQL_LONG = 3
MySQL_FLOAT = 4
MySQL_DOUBLE = 5
MySQL_NULL = 6
MySQL_TIMESTAMP = 7
MySQL_LONGLONG = 8
MySQL_INT24 = 9
MySQL_DATE = 10
MySQL_TIME = 11
MySQL_DATETIME = 12
MySQL_YEAR = 13
MySQL_NEWDATE = 14
MySQL_VARCHAR = 15
MySQL_BIT = 16
MySQL_NEWDECIMAL = 246
MySQL_ENUM = 247
MySQL_SET = 248
MySQL_TINY_BLOB = 249
MySQL_MEDIUM_BLOB = 250
MySQL_LONG_BLOB = 251
MySQL_BLOB = 252
MySQL_VAR_STRING = 253
MySQL_STRING = 254
MySQL_GEOMETRY = 255


"""Translation methods for MySQL field codes into the SQL data types(qal.sql.types):
return ['integer', 'string', 'string(255)', 'string(3000)', 'float', 'serial', 'timestamp', 'boolean']

Arguably, they should be declared in qal.dal.types instead.
"""


def mysql_type_to_SQL_type(_type_code):
    """
    Convert the internal MySQL type to the SQL type
    :param _type_code: A MySQL field constant
    """
    if _type_code in (
        MySQL_VARCHAR, 
        MySQL_VAR_STRING,
        MySQL_STRING
    ):
        return 'string'
    elif _type_code in  (
        MySQL_TINY_BLOB, 
        MySQL_MEDIUM_BLOB,
        MySQL_LONG_BLOB,
        MySQL_BLOB,
        ):
        return "blob"    
    elif _type_code in (
        MySQL_DECIMAL,
        MySQL_FLOAT,
        MySQL_DOUBLE,
        MySQL_NEWDECIMAL
        ):        
        return "float"
    elif _type_code in (
        MySQL_TINY,
        MySQL_SHORT,
        MySQL_LONG,
        MySQL_LONGLONG,
        MySQL_INT24,
        MySQL_BIT
        ):
        return "integer"    
    elif _type_code in (
        MySQL_DATE,
        MySQL_TIME,
        MySQL_DATETIME,
        MySQL_YEAR,
        MySQL_NEWDATE,
        MySQL_TIMESTAMP):
        return "timestamp"
    else:
        raise Exception("mysql_type_to_SQL_type: _type_code \"" + str(_type_code) + "\"not supported") 
    
def python_type_to_SQL_type(_python_type):
    """
    Convert a python data type to ab SQL type.
    :param _python_type: A Python internal type
    """
    if (_python_type == str):
        return 'string'
    elif (_python_type == bytes):
        return "blob"    
    elif (_python_type == float):
        return "float"
    elif (_python_type == int):
        return "integer"    
    elif (_python_type == datetime):
        return "datetime"
    elif (_python_type == bool):
        return "boolean"
    else:
        raise Exception("python_type_to_SQL_type: _type_code \"" + str(_python_type) + "\"not supported")
    
        
def parse_description(_descriptions, _db_type):
    """Convert field descriptions to field name- and field type-lists.
    :param _descriptions: A list of descriptions
    :param _db_type: The database type
    """
    _field_names = []
    _field_types = []
    
    for _column in (_descriptions):
        _field_names.append(_column[0])
        if _db_type == DB_MYSQL:
            _field_types.append(mysql_type_to_SQL_type(_column[1]))
        else:
            _field_types.append(_column[1])
    
    return _field_names, _field_types            
    



"""
    ****************************************
    qal.common - Commonly used functionality
    ****************************************

    This package gathers miscellaneous functionality that is used around QAL and the users of QAL. 
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""


"""
Created on Sep 14, 2012

@author: Nicklas Boerjesson
"""
from urllib.parse import quote
from datetime import datetime


from qal.dal.types import DB_DB2,DB_ORACLE, DB_POSTGRESQL
from qal.sql.utils import db_specific_object_reference
from qal.common.strings import empty_if_none
from qal.common.diff import compare

DATASET_LOGLEVEL_NONE = 0
DATASET_LOGLEVEL_LOW = 1
DATASET_LOGLEVEL_MEDIUM = 2
DATASET_LOGLEVEL_DETAIL = 3
DATASET_LOGLEVEL_ALL = 4


class CustomDataset(object):
    """This is the base class for all (external) data sets in QAL-
    Note: The fields are named like this to not appear as parameters in automatic generators like sql_xml.
    TODO: Field names should be considered captions, not references, and should not be used when looping columns.
    """

    field_names = None
    """These are, if applicable, the field names of the resulting dataset"""
    field_types = None
    """The data types of the fields"""
    data_table = None
    """The actual data, in the form of a two-dimensional array"""
    _base_path = None
    """The base path for all relative references in resources. Inferred from the locations of resource files."""

    
    _log_level = DATASET_LOGLEVEL_MEDIUM


    def __init__(self):
        """Constructor""" 
        self.field_names = []
        self.field_types = []
        self.data_table = []
        self._base_path = None
        self._log = [] 
        self._log_level = DATASET_LOGLEVEL_MEDIUM
        
    def cast_text_to_type(self, _text, _field_idx):
        try:
            if self.field_types is None:
                return _text
            elif self.field_types[_field_idx] == "integer":
                return int(_text)
            elif self.field_types[_field_idx] == "float":
                return float(_text)
            else:
                return _text
        except Exception as e:
            raise Exception("cast_text_to_type raised an error for \"" + _text +"\": " + str(e) )
        
    def log_update_row(self, _row_key, _old_row, _new_row, _comment = None):
        if self._log_level >= DATASET_LOGLEVEL_DETAIL:
            _field_diffs = []
            for _field_idx in range(len(self.field_names)):
                if _old_row[_field_idx] != _new_row[_field_idx]:
                    _field_diffs.append(self.field_names[_field_idx] + " : " + quote(str(_old_row[_field_idx])) + " =>" + quote(str(_new_row[_field_idx]))) 
                
            self._log.append(self.__class__.__name__ + ".update;"+quote(str(_row_key)) + ";"+";" +"|".join(_field_diffs) + empty_if_none(";"+ str(_comment), _comment))

    
    def log_insert(self, _row_key, _row_data, _comment = None):
        if self._log_level >= DATASET_LOGLEVEL_DETAIL:
            self._log.append(self.__class__.__name__ + ".insert;"+quote(str(_row_key)) + ";"+quote(str(_row_data)) + empty_if_none(";"+ str(_comment), _comment))

    def log_delete(self, _row_key, _row_data, _comment = None):
        if self._log_level >= DATASET_LOGLEVEL_DETAIL:
            self._log.append(self.__class__.__name__ + ".delete;"+quote(str(_row_key)) + ";"+quote(str(_row_data)) + empty_if_none(";"+ str(_comment), _comment))

    def log_save(self, _filename, _comment = None):
        if self._log_level >= DATASET_LOGLEVEL_LOW:
            self._log.append(self.__class__.__name__ + ".save;"+quote(str(_filename)) + ";" + str(datetime.now().isoformat()) + empty_if_none(";"+ str(_comment), _comment))
                
    def log_load(self, _filename, _comment = None):
        if self._log_level >= DATASET_LOGLEVEL_LOW:
            self._log.append(self.__class__.__name__ + ".load;"+quote(str(_filename)) + ";" + str(datetime.now().isoformat()) + empty_if_none(";"+ str(_comment), _comment))

    def read_resource_settings(self, _resource):
        """Load the data"""
        raise Exception('CustomDataset.read_resource_settings is not implemented in class: ' + self.__class__.__name__)

    def write_resource_settings(self, _resource):
        """Load the data"""
        raise Exception('CustomDataset.write_resource_settings is not implemented in class: ' + self.__class__.__name__)

    def load(self):
        """Load the data"""
        raise Exception('CustomDataset.load is not implemented in class: ' + self.__class__.__name__)
        pass
    
    def save(self):
        """Save the data to the underlying target."""
        raise Exception('CustomDataset.save is not implemented in class: ' + self.__class__.__name__)

    def clear_log(self):
        self._log = []
    
    def _structure_insert_row(self, _row_idx, _row_data, _commit=None,_no_logging = False):
        """Inserts a row at _row_idx in the self.data_table, containing _row_data\
        Commonly overridden by subclasses"""
        if _no_logging == False:
            self.log_insert(_row_idx, _row_data)

        self.data_table.insert(_row_idx,_row_data)

    def _structure_update_row(self, _row_idx, _row_data, _commit=None, _no_logging = False):
        """Updates the row at _row_idx in the self.data_table with _row_data.\
        Commonly overridden by subclasses"""
        if _no_logging == False:
            self.log_update_row(_row_idx, self.data_table[_row_idx], _row_data)
        self.data_table[_row_idx] = _row_data
    def _structure_delete_row(self, _row_idx, _commit=None, _no_logging = False):
        """Deletes a row at _row_idx from the self.data_table.\
        Commonly overridden by subclasses"""
        if _no_logging == False:
            self.log_delete(_row_idx, self.data_table[_row_idx])
        self.data_table.pop(_row_idx)
        
        
    def _structure_init(self, _dataset):
        """Initialize underlying structure before applying data to it.\
        Overridden by subclasses."""
        pass

        
    def _structure_apply_merge(self, _insert, _update, _delete, _sorted_dest, _commit=True):
        
        #print("_insert: " + str(_insert))
        #print("Before apply: " + str(_sorted_dest))
        if len(self._structure_key_fields) >= len(self.field_names):
            raise Exception("Error in RDBMSDataset.structure_init: Updates cannot be made when there are no non-key fields.")

        if _insert:
            _insert_idx = len(_insert) - 1
        else:
            _insert_idx = -1
            
        if _delete:
            _delete_idx = len(_delete) - 1
        else:
            _delete_idx = -1
            
        if _update:
            _update_idx = len(_update) - 1
        else:
            _update_idx = -1

        
        self.data_table = _sorted_dest
        if len(self.data_table) > 0:
            # Loop the sorted destination dataset backwards and apply changes
            for _curr_row_idx in range(len(self.data_table), -1, -1):

                # Make deletes first, as they displace everything after them
                while _delete_idx > -1 and _delete[_delete_idx][1] == _curr_row_idx:
                    self._structure_delete_row(_curr_row_idx, _commit=_commit)
                    _delete_idx-= 1
                # Make updates
                while _update_idx > -1 and _update[_update_idx][1] == _curr_row_idx:
                    self._structure_update_row(_curr_row_idx, _update[_update_idx][2], _commit=_commit)
                    _update_idx-= 1

                # Make inserts
                while _insert_idx > -1 and _insert[_insert_idx][1] == _curr_row_idx:
                    self._structure_insert_row(_curr_row_idx, _insert[_insert_idx][2], _commit=_commit)
                    _insert_idx-= 1
        else:
            # Make inserts
            while _insert_idx > -1:
                self._structure_insert_row(0, _insert[_insert_idx][2], _commit=_commit)
                _insert_idx-= 1

        #print("After apply:  " + str(_sorted_dest))            
        return self.data_table        
    
    def apply_new_data(self, _new_data_table, _key_fields, _insert=None, _update=None, _delete=None, _commit=True):
        """This function applies a new data table unto the existing, matches are made using the key fields.
      
        
        :parameter 2D-list _new_data_table: A two-dimensional list contains the data. Must match the existing column-wise.
        :parameter list _key_fields: An array with the indices of the fields that should be used to match source rows to destination rows. 
        
        -- note:
            If there are different data types in the _new_data_table columns and the existing dataset.data_table,
            they will be considered different and be updated.
            It is also possible that the keys will not match. So cast these before applying.
        """
         
        self._structure_key_fields = _key_fields
        
        # Some data sets needs to initialize the underlying structure to work, and work efficiently
        self._structure_init(_new_data_table)

        if len(self._structure_key_fields) > 0:
            #  If there are defined keys Compare the source and destination table to generate difference sets
            _deletes, _inserts, _updates, _dest_sorted = compare(
                                                              _left = _new_data_table,
                                                              _right = self.data_table,
                                                              _key_columns = _key_fields,
                                                              _full = True)
        else:
            # If there are no keys, just make inserts of all rows
            _inserts = [[0, 0, _new_data_table[_curr_idx]] for _curr_idx in range(len(_new_data_table) - 1, -1, -1)]
            # There will be no deletes or updates
            _deletes = _updates = []
            # We have no indexes to sort with, keep current sort order
            _dest_sorted = []

        # Merge the data into the structure. 
        # Note: in RDBMSDataset, this currently means writing to the underlying database, since there is no in-memory structure.

        # Clear out all changes that should not be made
        if _delete is None or _delete==False:
            _deletes=None
        if _insert is None or _insert==False:
            _inserts=None
        if _update is None or _update==False:
            _updates=None

        self.data_table = self._structure_apply_merge(_inserts, _updates, _deletes, _dest_sorted, _commit=_commit)
        
        return self.data_table, _deletes, _inserts, _updates
                            
                
        
    
    def as_sql(self, _db_type):
        """Generate SQL
        Either through a union SQL or into a temp table."""
        _result = []
        
        _add_field_names = (len(self.data_table) > 0) and (len(self.field_names) == len(self.data_table[0]))
        
        
        # TODO: If is i a large number of rows, make an insert into a temp table instead.
        for _row in self.data_table:
            _curr_row = []
            for _col_idx in range(len(_row)):
                _str_col = ''
                _col = _row[_col_idx]
                if (_col.lower() == ''):
                    _str_col = 'NULL'                    
                elif (_col.lower() in ['true', 'false']):
                    if (_col.lower() == 'true'):
                        if (_db_type == DB_POSTGRESQL):
                            _str_col = 'TRUE'
                        else:
                            _str_col = '\'1\''
                    else:
                        if (_db_type == DB_POSTGRESQL):
                            _str_col = 'FALSE'
                        else:
                            _str_col = '\'0\''
                else:
                    _str_col = _col 
                            
                if (_add_field_names):                    
                    _curr_row.append(_str_col + ' AS ' + db_specific_object_reference(self.field_names[_col_idx], _db_type))
                else:
                    _curr_row.append(_str_col)
            
            _add_field_names = False
                
            if (_db_type == DB_DB2):
                _result.append("SELECT " + ",".join(_curr_row) +' FROM sysibm.sysdummy1')
            elif (_db_type == DB_ORACLE):
                _result.append("SELECT " + ",".join(_curr_row) +' FROM DUAL')
            else:
                _result.append("SELECT " + ",".join(_curr_row))
            
                    
        return str("\nUNION\n".join(_result)) 




"""
    ******************************
    qal.dataset - Data set classes
    ******************************
    The dataset package contains different data sets, all inheriting from the :py:class:qal.dataset.custom.custom class.
    They provide a generalized way of reading and writing to a multitude of data sources.
    Currenctly, there are: flat file, matrix, XPath(XML/HTML/XHTML) and spreadsheet(Excel-formats/.odt)
    
    :copyright: Copyright 2010-2014 by Nicklas Boerjesson
    :license: BSD, see LICENSE for details.
"""


'''
Created on Oct 2, 2012

@author: Nicklas Boerjesson
'''
from qal.dal.types import DB_POSTGRESQL, DB_MYSQL, DB_ORACLE, DB_DB2, DB_SQLSERVER 
from qal.common.listhelper import unenumerate


def check_for_param_content(_value):
    if (_value[0:8] == '::Param='):
        return True
    else:
        return False

def none_as_sql(_value, _db_type, _none_value = '', _error = None):
    """Raises an error or returns a value if blank"""
    if _value is not None:
        return _value.as_sql(_db_type)
    else:
        if _error is not None:
            raise Exception(_error)
        else:
            return _none_value
        
def check_not_null(_classname, _items):
    for _item in _items:
        if not _item[0]:
            raise Exception("Error in " + _classname + ", " + _item[1] + " is not set.")


def error_on_blank(_value, _error):
    """Raises an error if blank"""
    if _value not in [None, '']:
        return _value
    else:
        raise Exception(_error)
    
def citate(AValue, _db_type):
    """Adds citations when db_type is DB_POSTGRESQL"""
    if _db_type in (DB_POSTGRESQL, DB_DB2, DB_ORACLE):
        return '"' + str(AValue) + '"'
    else:
        return AValue
    
def comma_separate(_list, _db_type):
    """Convert a list of possible ParameterBase-descendants."""
    result = ''
    for currItem in _list:
        if result != '': result = result + ","
        if hasattr(currItem, 'as_sql'):
            result+= currItem.as_sql(_db_type)
        else:
            result+= citate(currItem, _db_type)

    return result

def tabular_source_to_sql(_tabular_source, _db_type):
    pass
    

def parenthesise(_value):
    """Adds a parenthesis around a values."""
    return '(' + _value + ')'

def oracle_add_escape(_value, _escape_character):
    """Add the oracle escape ke_delimiteryword if nessesary."""
    if _escape_character not in ('', None):
        return _value + ' ESCAPE \'' + _escape_character + '\''
    else:
        return _value  

def oracle_create_auto_increment(_table,_column):
    """Work around Oracle's silly unwillingness to support auto increment."""
    result = list()
    _new_seq = (_table.name+'_' + _column.name +'_seq')[0:29]
    result.append('')
    result[0]= 'CREATE SEQUENCE ' + _new_seq + _table._row_separator  
    result[0]+= 'start with 1' + _table._row_separator  
    result[0]+= 'increment by 1 ' + _table._row_separator  
    result[0]+= 'nomaxvalue';  
    result.append('')
    result[1]= 'CREATE TRIGGER ' + ('tr_' + _table.name + '_' + _column.name)[0:29]+ _table._row_separator  
    result[1]+= 'BEFORE INSERT ON "' + _table.name + '"'+ _table._row_separator  
    result[1]+= 'FOR EACH ROW WHEN (new."' + _column.name +'" IS NULL) BEGIN'+ _table._row_separator  
    result[1]+= 'SELECT '+ _new_seq +'.nextval INTO :new."' + _column.name +'" FROM dual;'+ _table._row_separator   
    result[1]+= 'END;'       
    return result 
      
def add_operator(_index, _operator):
    """Adds a an operator if index > 0, used in loops when making conditions."""
    if _index > 0:
        return ' ' + _operator + ' '
    else:
        return ''

def add_comma(_index, _value):
    """Adds a comma if index > 0, used in loops when making lists."""
    if _index > 0:
        return ', ' + _value
    else:
        return _value
def add_comma_rs(_index, _value, _row_separator):
    """Adds a comma and row separator if index > 0, used in loops when making lists of references."""    
    if _index > 0:
        return ',' + _row_separator + _value
    else:
        return _value    

# Some database functions
def curr_user(_db_type):
    """Returns a database-type specific (see dal_types) way of getting the current user.""" 
    if   _db_type == DB_MYSQL:
        return '"Not supported by MySQL"'
    elif _db_type == DB_POSTGRESQL:
        return 'session_user'
    elif _db_type == DB_ORACLE:
        return 'USER'
    elif _db_type == DB_DB2:
        return 'SESSION_USER'
    elif _db_type == DB_SQLSERVER:
        return 'SUSER_SNAME()'
    
def curr_datetime(_db_type):
    """Returns a database-type specific (see dal_types) way of getting the current date time."""
    if   _db_type == DB_MYSQL:
        return 'CURRENT_TIMESTAMP'
    elif _db_type == DB_POSTGRESQL:
        return 'current_timestamp'
    elif _db_type == DB_ORACLE:
        return 'CURRENT_TIMESTAMP'
    elif _db_type == DB_DB2:
        return 'CURRENT_TIMESTAMP'
    elif _db_type == DB_SQLSERVER:
        return 'GETDATE()'      
        
def db_specific_object_reference(_value, _db_type):
    """Qualifies an object reference in a database-type specific (see dal_types) way."""

    if _db_type == DB_MYSQL:
        return "`" + _value + "`"
    elif _db_type == DB_POSTGRESQL:
        return '"' + _value + '"'
    elif _db_type == DB_ORACLE:
        return '"' + str(_value)[0:30] + '"'
    elif _db_type == DB_DB2:
        return '"' + _value + '"'
    elif _db_type == DB_SQLSERVER:
        return '[' + _value + ']'

    else:
        raise Exception("db_specific_object_reference: Invalid database type.")    
def db_specific_datatype_parse_length(_value): 
    if (_value.find('(') > -1):
        _strlength = _value.split('(', 1)[1].rsplit(')', 1)[0]
        if ((_strlength.lower() != '') and (_strlength.isdigit() == False)):
            raise Exception("db_specific_datatype: Invalid syntax for datatype length:" + _value)
    else:
        _strlength = ''
            
    return _strlength    
    
def db_specific_datatype(value, DB):
    """Converts general DAL datatypes(as defined in sql_types) database-type specific (see dal_types) representations."""

    # Database order: [DB_MYSQL, DB_POSTGRESQL, DB_ORACLE, DB_DB2, DB_SQLSERVER]

    result = ''
    if (value.lower() == "integer"):
        result = unenumerate(['INTEGER', 'INTEGER', 'NUMBER', 'INT', 'int'], DB)
    elif (value[:6].lower() == "string" or value[:7].lower() == "varchar"):
        strlength = db_specific_datatype_parse_length(value)
        if (strlength.lower() != ''):
            result = unenumerate(['VARCHAR', 'VARCHAR', 'VARCHAR2', 'VARCHAR', 'varchar'], DB)
            result+= parenthesise(db_specific_datatype_parse_length(value))
        else:
            # @attention: For some reason, DB2 and Oracle doesn't support unspecified column 
            # lengths unless using BLOBs and cumbersome select constructs and casts.
            # So DB2 will get 3100(max is 32704, but 4000 is the default page size for 
            # the table space so 3100 allows for the unspecified column to be at least usable.)
            # Oracle is 4000 (limitations confirmed as of dec. 2012). 
            result = unenumerate(['TEXT', 'VARCHAR', 'VARCHAR2(4000)', 'VARCHAR(3100)', 'varchar(max)'], DB)
            
    elif (value.lower() == "float"):
        result = unenumerate(['DOUBLE', 'double precision', 'FLOAT', 'Double', 'float'], DB)    
    
    elif (value.lower() == "serial"):
        result = unenumerate(['INTEGER AUTO_INCREMENT', 'serial', 'integer','INT GENERATED ALWAYS AS IDENTITY', 'int IDENTITY(1,1)'], DB)
    elif (value.lower() == "timestamp"):
        result = unenumerate(['TIMESTAMP', 'TIMESTAMP', 'TIMESTAMP','TIMESTAMP', 'DATETIME'], DB)
    elif (value.lower() == "datetime"):
        result = unenumerate(['DATETIME', 'TIMESTAMP', 'DATE','TIMESTAMP', 'DATETIME'], DB)
    elif (value.lower() == "boolean"):
        result = unenumerate(['BOOL', 'BOOLEAN', 'NUMERIC(1)','DECIMAL(1)', 'BIT'], DB)

        #elif (value.lower() == "timestamp"):
    #    result = unenumerate(['TIMESTAMP DEFAULT CURRENT_TIMESTAMP', 'DATETIME DEFAULT(NOW())', 'DATETIME DEFAULT(NOW())','TIMESTAMP', 'DATETIME DEFAULT(NOW())'], DB)
    else:
        result = value        
    return result

def db_specific_operator(_operator, _db_type):
    """Makes PostgreSQL's LIKE operator case insensitive by uppercasing it.""" 
    if str.lower(_operator) == 'like' and _db_type == DB_POSTGRESQL:
        return 'ILIKE'
    else:
        return _operator
  
def make_operator(_operator, _double_pipe_c):
    """Handles concatenation operators."""
    if _operator == 'C':
        if _double_pipe_c == True:
            return '||'
        else:
            return '+'
    else:
        return _operator

def make_function(_name, _value):
    """Assembles an SQL function call."""
    return _name + parenthesise(_value)


def handle_temp_table_ref(_identifier, _db_type):
    """Removes the "#" in SQL Server temporary table references. """
    if len(_identifier) > 0 and _identifier[0] == "#" and  _db_type != DB_SQLSERVER:
        return _identifier[1:]
    else:
        return _identifier
    
def datatype_to_parameter(_db_type, _datatype):
    """Converts a python data type to the database-driver appropriate parameter substitution string"""
    if _db_type == DB_MYSQL:
        return "%s"
    elif (_datatype[:6].lower() == "string" or _datatype[:7].lower() == "varchar" or _datatype in ["timestamp", "datetime"]):
        return "%s"
    elif (_datatype in ["float", "integer"]):
        return "%d"
    else:
        raise Exception("datatype_to_parameter, unsupported data_type: " + str(_datatype))            
    